/**
 ** bpp
 ** Bidirectional parallel port driver.
 **/

#include <sys/types.h>
#include <sys/systm.h>		/* nodev, nulldev, timeout */
#include <sys/conf.h>		/* D_NEW */
#include <sys/modctl.h>		/* struct modldrv */
#include <sys/fcntl.h>		/* O_XXXX */
#include <sys/file.h>		/* FWRITE */
#include <sys/stat.h>		/* S_IFCHR */
#include <sys/errno.h>
#include <sys/open.h>		/* OTYP_CHR */

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include <sys/bpp_io.h>
#include <sys/bpp_var.h>
#include <sys/bpp_reg.h>

static void check_for_active_pins(int instance);
static void bpp_minphys(struct buf *bp);
static void bpp_transfer_failed(int instance);
static void set_dss_dsw(int instance, int mode);
static short check_write_params(struct bpp_transfer_parms *, int, int);
static short check_read_params(struct bpp_transfer_parms *, int, int);
static void read_outpins(int instance, int mode, int handshake);

static int bpp_open(dev_t *devp, int flag, int otyp, cred_t *);
static int bpp_close(dev_t dev, int flag, int otyp, cred_t *);
static int bpp_read(dev_t dev, struct uio *uio_p, cred_t *);
static int bpp_write(dev_t dev, struct uio *uio_p, cred_t *);
static int bpp_ioctl(dev_t, int, int, int, cred_t *, int *);
static int bpp_strategy(struct buf *bp);
static u_int bpp_intr(caddr_t instance);
static void bpp_transfer_timeout(caddr_t arg);

static int bpp_getinfo(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int bpp_identify(dev_info_t *);
static int bpp_attach(dev_info_t *, ddi_attach_cmd_t);
static int bpp_detach(dev_info_t *, ddi_detach_cmd_t);

static struct bpp_transfer_parms bpp_default_transfer_parms = {	/* :0 */
	BPP_ACK_BUSY_HS,	/* :00 read_handshake */
	1000,			/* :04 read_setup_time, nsec */
	1000,			/* :08 read_strobe_width, nsec */
	60,			/* :12 read_timeout, usec */
	BPP_ACK_HS,		/* :16 write_handshake */
	1000,			/* :20 write_setup_time, nsec */
	1000,			/* :24 write_strobe_width, nsec */
	60			/* :28 write_timeout, usec */
};

static struct bpp_pins bpp_default_pins;		/* 0x20==32 */
static struct bpp_error_status bpp_default_error_stat;	/* 0x22==34 */
static int sbus_clock;					/* 0x28==40 */
static int sbus_cycle;					/* 0x2c==44 */
static struct ddi_dma_lim bpp_limits = {		/* 0x30==48 */
	0,		/* dlim_addr_lo */
	0xffffffff,	/* dlim_addr_hi */
	0x00ffffff,	/* dlim_cntr_max */
/* data 0x3c: */  0x1f,	/* dlim_burstsizes */
/* data 0x40: */     1,	/* dlim_minxfer */
	0		/* dlim_dmaspeed */
};

static struct cb_ops bpp_cb_ops = {
	bpp_open,
	bpp_close,
	nodev,
	nodev,
	nodev,
	bpp_read,
	bpp_write,
	bpp_ioctl,
	nodev,
	nulldev,
	nulldev,
	nochpoll, 
	ddi_prop_op,
	0,			/* STREAMS or not? */
	(D_NEW|D_MP),
	CB_REV,
	nodev,
	nodev
};

static struct dev_ops bpp_ops = {	/* :90 */
	DEVO_REV,		/* :90 */
	0,			/* :94 */
	bpp_getinfo,		/* :98 */
	bpp_identify,
	nulldev,		/* :a0 */
	bpp_attach,
	bpp_detach,		/* :b0 */
	nodev,
	&bpp_cb_ops		/* :b8 */
};

static struct modldrv modldrv = {	/* :bc */
	&mod_driverops,
	"bidir parallel port 2.0 @(#) 91/15/11",
	&bpp_ops
};

static struct modlinkage modlinkage = {		/* :c8 */
	MODREV_1,
	{  &modldrv,  0, 0, 0 }
};

static void *bpp_state_head;		/* off=0x0 */

static int bpp_identify(dev_info_t *devinfo)
{
	char *dev_name;
	dev_name = ddi_get_name(devinfo);

	if (strcmp(dev_name, BPP_PROM_NAME) == 0) {
		return DDI_IDENTIFIED;
	} else {
		return DDI_NOT_IDENTIFIED;
	}
}

static int bpp_attach(dev_info_t *devinfo, ddi_attach_cmd_t cmd)
{
	register struct bpp_unit *lp;
	register volatile struct bpp_regs *regs;
	int ins;
	char sbuff[16];
	int burstsize;
	int dum;

	ins = ddi_get_instance(devinfo);
	if (cmd != DDI_ATTACH) {
		if (cmd != DDI_RESUME) return -1;
		lp = (struct bpp_unit *) ddi_get_driver_private(devinfo);
		if (lp == 0) return -1;

		mutex_enter(&lp->bpp_mutex);
		if (lp->unit_suspended == 0) {
			mutex_exit(&lp->bpp_mutex);
			return -1;
		}
		lp->unit_suspended = 0;
		mutex_exit(&lp->bpp_mutex);
		regs = lp->bpp_regs_p;

	} else {

		if (ddi_slaveonly(devinfo) == DDI_SUCCESS) {
			cmn_err(CE_NOTE, "bpp unit %d: NOT used - SBus slot is slave only", ins);
			return -1;
		}
		if (ddi_soft_state_zalloc(bpp_state_head, ins) != 0) return -1;
		lp = ddi_get_soft_state(bpp_state_head, ins);
		ddi_set_driver_private(devinfo, (caddr_t)lp);

		if (ddi_add_intr(devinfo, 0, &lp->bpp_block_cookie, 0, bpp_intr, (caddr_t)ins) != 0) {
			cmn_err(CE_NOTE, "bpp_attach unit %d: cannot add interrupt!", ins);
			ddi_soft_state_free(bpp_state_head, ins);
			return -1;
		}
		cv_init(&lp->wr_cv, "multi-write synchron", CV_DRIVER, NULL);
		mutex_init(&lp->bpp_mutex, "bpp mutex", 4, lp->bpp_block_cookie);
		lp->dip = devinfo;
		lp->transfer_parms = bpp_default_transfer_parms;

		lp->pins = bpp_default_pins;
		lp->error_stat = bpp_default_error_stat;

		lp->timeouts = NO_TIMEOUTS;
		lp->bpp_transfer_timeout_ident = -1;
		lp->bpp_fakeout_timeout_ident = -1;

		sbus_clock = ddi_getprop(DDI_DEV_T_ANY, devinfo, 0, "clock-frequency", 1000000)/1000000;

		if (sbus_clock < 10 || sbus_clock > 25) {
			cmn_err(CE_NOTE, "SBus clock frequency %x out of range", sbus_clock);
			/* mutex_destroy() is forgotten here */
			ddi_remove_intr(lp->dip,0,lp->bpp_block_cookie);
			ddi_soft_state_free(bpp_state_head, ins);
			return -1;
		}
		sbus_cycle = 1000/sbus_clock;
		lp->sbus_clock_cycle = sbus_cycle;

		if (ddi_map_regs(devinfo, 0, (caddr_t *) &lp->bpp_regs_p, 0, 28) != 0) {
			cmn_err(CE_NOTE, "bpp_attach unit %d: ddi_map_regs failed!", ins);
			mutex_destroy(&lp->bpp_mutex);
			ddi_remove_intr(lp->dip,0,lp->bpp_block_cookie);
			ddi_soft_state_free(bpp_state_head, ins);
			return -1;
		}
		regs = lp->bpp_regs_p;

		if (check_bpp_registers(ins) != 0) {
			cmn_err(CE_NOTE, "bpp_attach unit %d: register check failed!", ins);
			ddi_unmap_regs(devinfo, 0, (caddr_t *) &lp->bpp_regs_p, 0, 28);
			mutex_destroy(&lp->bpp_mutex);
			ddi_remove_intr(lp->dip, 0, lp->bpp_block_cookie);
			ddi_soft_state_free(bpp_state_head, ins);
			return -1;
		}

		sprintf(sbuff, "bpp%d", ins);
		if (ddi_create_minor_node(devinfo, sbuff, S_IFCHR, ins, 0, 0) == -1) {
			ddi_remove_minor_node(devinfo, 0);
			cmn_err(CE_NOTE, "ddi_create_minor_node failed for unit %d\n", ins);
			ddi_unmap_regs(devinfo, 0, (caddr_t *) &lp->bpp_regs_p, 0, 28);
			mutex_destroy(&lp->bpp_mutex);
			ddi_remove_intr(lp->dip, 0, lp->bpp_block_cookie);
			ddi_soft_state_free(bpp_state_head, ins);
			return -1;
		}

		ddi_report_dev(devinfo);

	}

	burstsize = ddi_getprop(-2, devinfo, 0, "burst-sizes", -1);

	/* ????????? */
	if ((regs->dma_csr & BPP_DEVICE_ID_MASK) != 0xa0000000) {
		dum = regs->dma_csr;	/* ??? */
		if (burstsize == 255 || (burstsize & 0x10) != 0) {
			dum = regs->dma_csr;	/* ??? */
		} else if ((burstsize & 0x20) != 0) {
			regs->dma_csr |= 0x40000;
		} else if ((burstsize & 0x10) != 0) {
			dum = regs->dma_csr;
		}
	} else if ((regs->dma_csr & BPP_DEVICE_ID_MASK) != 0x40000000) {
		dum = regs->dma_csr;
	}

	regs->dma_csr |= BPP_RESET_BPP;
	regs->dma_csr &= ~BPP_RESET_BPP;
	regs->dma_csr |= BPP_TC_INTR_DISABLE;
	regs->int_cntl |= BPP_ERR_IRP;
	regs->int_cntl |= BPP_SLCT_IRP;
	regs->int_cntl &= ~BPP_PE_IRP;
	regs->int_cntl |= BPP_BUSY_IRP;
	regs->int_cntl |= BPP_ALL_IRQS;
	lp->bpp_regs_p->dma_csr |= BPP_INT_EN;

	lp->versatec_allowed = !(lp->bpp_regs_p->op_config & BPP_VERSATEC_INTERLOCK);

	return 0;
}

static int
bpp_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg, void **resultp)
{
	struct bpp_unit *lp;

	switch (cmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if ((lp = ddi_get_soft_state(bpp_state_head, (int)arg)) == 0) {
			*resultp = 0;
			return -1;
		}
		mutex_enter(&lp->bpp_mutex);
		*resultp = lp->dip;
		mutex_exit(&lp->bpp_mutex);
		return 0;
	case DDI_INFO_DEVT2INSTANCE:
		*(int*)resultp = getminor((dev_t)arg);
		return 0;
	default:
		return -1;
	}
}

int _init()
{
	int error;
	error = ddi_soft_state_init(&bpp_state_head, sizeof(struct bpp_unit), 1);
	if (error == 0) {
		error = mod_install(&modlinkage);
		if (error != 0) {
			ddi_soft_state_fini(&bpp_state_head);
		}
	}
	return error;
}

int _info(struct modinfo *modinfop)
{
	return mod_info(&modlinkage, modinfop);
}

int _fini()
{
	int error;
	error = mod_remove(&modlinkage);
	if (error == 0) {
		ddi_soft_state_fini(&bpp_state_head);
	}
	return error;
}

/*
 */
static int
bpp_detach(dev_info_t *devinfo, ddi_detach_cmd_t cmd)
{
	struct bpp_unit *lp;
	volatile struct bpp_regs *regs;
	int ins;

	ins = ddi_get_instance(devinfo);

	if (cmd != 0) {
		if (cmd != 1) return -1;
		if ((lp = (struct bpp_unit *) ddi_get_driver_private(devinfo)) == 0) return -1;
		mutex_enter(&lp->bpp_mutex);
		if (lp->unit_suspended) {
			mutex_exit(&lp->bpp_mutex);
			return -1;
		}
		lp->unit_suspended = 1;
		mutex_exit(&lp->bpp_mutex);
		while (lp->timeouts) {}
		return 0;
	}
	lp = ddi_get_soft_state(bpp_state_head, ins);
	regs = lp->bpp_regs_p;
	if (regs->dma_csr & BPP_ENABLE_DMA) {
		cmn_err(CE_NOTE, "ERROR: bpp unload of unit %d while DMA active!", ins);
		regs->dma_csr &= ~BPP_ENABLE_DMA;
		regs->dma_csr &= ~BPP_ENABLE_BCNT;
		regs->op_config |= BPP_SRST;
		regs->op_config &= ~BPP_SRST;
		regs->op_config |= BPP_FLUSH;
	}
	regs->dma_csr |= BPP_TC_INTR_DISABLE;
	regs->int_cntl &= ~(BPP_PE_IRQ_EN|BPP_SLCT_IRQ_EN|BPP_ERR_IRQ_EN);
	regs->dma_csr &= ~BPP_INT_EN;

	ddi_remove_minor_node(devinfo, 0);
	ddi_unmap_regs(lp->dip, 0, (caddr_t *)&lp->bpp_regs_p, 0, sizeof(struct bpp_regs));
	ddi_remove_intr(lp->dip, 0, lp->bpp_block_cookie);
	mutex_destroy(&lp->bpp_mutex);
	ddi_soft_state_free(bpp_state_head, ins);
	return 0;
}

static int bpp_open(dev_t *devp, int flag, int otyp, cred_t *cred_p)
{
	int ins = getminor(*devp);
	struct bpp_unit *lp = ddi_get_soft_state(bpp_state_head, ins);
	int error = 0;

	if (otyp != OTYP_CHR) {
		cmn_err(CE_NOTE, "bpp%d attempted open as non-character device!", ins);
		return EINVAL;
	}
	if (lp == 0) {
		cmn_err(CE_NOTE, "bpp%d unit pointer is NULL!", ins);
		return ENXIO;
	}
	mutex_enter(&lp->bpp_mutex);
	if (lp->unit_open) {
		error = EBUSY;
		goto cleanup;
	}
	lp->unit_open = 1;
	lp->transfer_parms = bpp_default_transfer_parms;
	lp->pins = bpp_default_pins;
	lp->openflags = flag;
cleanup:
	mutex_exit(&lp->bpp_mutex);
	return error;
}

static int bpp_close(dev_t dev, int flag, int otyp, cred_t *cred_p)
{
	struct bpp_unit *lp = ddi_get_soft_state(bpp_state_head, getminor(dev));

	mutex_enter(&lp->bpp_mutex);
	if (lp->timeouts != NO_TIMEOUTS) {
		if (lp->timeouts & TRANSFER_TIMEOUT) {
			untimeout(lp->bpp_transfer_timeout_ident);
		}
		if (lp->timeouts & FAKEOUT_TIMEOUT) {
			/* CE_CONT ?? */
	 		cmn_err(CE_CONT, "BOGUS fakeout timeout.\n");
		}
	}
	lp->unit_open = 0;
	lp->timeouts = NO_TIMEOUTS;
	mutex_exit(&lp->bpp_mutex);
	return 0;
}

static int bpp_read(dev_t dev, struct uio *uio_p, cred_t *cred_p)
{
	static int int_dc = 1000;

	int ins;
	struct bpp_unit *lp;
	register volatile struct bpp_regs *regs;

	ins = getminor(dev);
	lp = ddi_get_soft_state(bpp_state_head, ins);

	/* Note an access to *lp before aquiring the mutex. Strange. */
	regs = lp->bpp_regs_p;

	if (lp->last_trans == write_trans) {
		drv_usecwait(int_dc);
	}

	mutex_enter(&lp->bpp_mutex);

	regs->op_config &= ~BPP_EN_MEM_CLR;

	switch (lp->transfer_parms.read_handshake) {
	case BPP_NO_HS:
		regs->op_config &= ~(BPP_BUSY_OP|BPP_ACK_OP);
		regs->op_config |= BPP_BUSY_BIDIR|BPP_DS_BIDIR;
		break;
	case BPP_ACK_HS:
		regs->op_config &= ~BPP_BUSY_OP;
		regs->op_config |= BPP_ACK_OP;
		regs->op_config |= (BPP_BUSY_BIDIR|BPP_DS_BIDIR|BPP_ACK_BIDIR);
		break;
	case BPP_BUSY_HS:
	case BPP_HSCAN_HS:
		regs->op_config |= BPP_BUSY_OP;
		regs->op_config &= ~BPP_ACK_OP;
		regs->op_config |= (BPP_BUSY_BIDIR|BPP_DS_BIDIR);
		break;
	case BPP_ACK_BUSY_HS:
		regs->op_config |= (BPP_ACK_OP|BPP_BUSY_OP);
		regs->op_config |= (BPP_BUSY_BIDIR|BPP_DS_BIDIR|BPP_ACK_BIDIR);
		break;
	case BPP_XSCAN_HS:
		regs->op_config &= ~BPP_BUSY_OP;
		regs->op_config |= BPP_ACK_OP;
		regs->op_config &= ~(BPP_DS_BIDIR|BPP_BUSY_BIDIR|BPP_ACK_BIDIR);
		break;
	case BPP_CLEAR_MEM:
		regs->op_config &= ~BPP_DMA_DATA;
		regs->op_config |= BPP_EN_MEM_CLR;
		break;
	case BPP_SET_MEM:
		regs->op_config |= BPP_DMA_DATA;
		regs->op_config |= BPP_EN_MEM_CLR;
		break;
	default:
		;
	}
	regs->trans_cntl |= BPP_DIRECTION;	/* read */
	set_dss_dsw(ins, 1);

	/** ???? Impossible condition, */
	if ((lp->openflags & O_WRONLY) != 0 && (lp->openflags & O_RDWR) != 0) {

		if (lp->transfer_parms.read_handshake == BPP_HSCAN_HS) {
			regs->out_pins |= BPP_AFX_PIN;
		} else {
			regs->out_pins |= BPP_SLCTIN_PIN;
		}
	}
	lp->last_trans = read_trans;
	mutex_exit(&lp->bpp_mutex);

	/* physio() returns signed short, which is converted. */
	return physio(bpp_strategy, 0, dev, B_READ, bpp_minphys, uio_p);
}

/*
 */
static int bpp_write(dev_t dev, struct uio *uio_p, cred_t *cred_p)
{
	int ins;
	struct bpp_unit *lp;
	register volatile struct bpp_regs *regs;
	int error;

	ins = getminor(dev);
	lp = ddi_get_soft_state(bpp_state_head, ins);
	regs = lp->bpp_regs_p;
	mutex_enter(&lp->bpp_mutex);

	lp->error_stat = bpp_default_error_stat;

	while ((lp->timeouts & TRANSFER_TIMEOUT) != 0) {
		if (cv_wait_sig(&lp->wr_cv, &lp->bpp_mutex) == 0) {
			mutex_exit(&lp->bpp_mutex);
			return EINTR;
		}
	}

	regs->int_cntl |= BPP_ERR_IRP;
	regs->int_cntl |= BPP_SLCT_IRP;
	regs->int_cntl &= ~(BPP_PE_IRP);
	regs->int_cntl |= BPP_PE_IRQ|BPP_SLCT_IRQ|BPP_ERR_IRQ;

	check_for_active_pins(ins);

	if ((lp->openflags & 0x03) == 3) {	/* Same "impossible" ???? */
		if (lp->transfer_parms.read_handshake == BPP_HSCAN_HS) {
			regs->out_pins &= ~BPP_AFX_PIN;
		} else {
			regs->out_pins &= ~BPP_SLCTIN_PIN;
		}

	} else {
		if ((lp->error_stat.pin_status & BPP_ALL_IN_PINS) != 0) {
			mutex_exit(&lp->bpp_mutex);
			error = ENXIO;
			goto clean_cv_and_return;
		}
	}

	regs->trans_cntl &= ~BPP_DIRECTION;
	regs->op_config &= ~BPP_EN_MEM_CLR;

	if (lp->transfer_parms.write_handshake == BPP_NO_HS) {
		regs->op_config &= ~(BPP_ACK_OP|BPP_BUSY_OP);
	} else if (lp->transfer_parms.write_handshake == BPP_ACK_HS) {
		regs->op_config &= ~BPP_BUSY_OP;
		regs->op_config |= BPP_ACK_OP;
	} else if (lp->transfer_parms.write_handshake == BPP_BUSY_HS) {
		regs->op_config |= BPP_BUSY_OP;
		regs->op_config &= ~BPP_ACK_OP;
	}
	regs->op_config &= ~(BPP_BUSY_BIDIR|BPP_ACK_BIDIR);

	set_dss_dsw(ins, 0);
	lp->last_trans = write_trans;
	mutex_exit(&lp->bpp_mutex);

	error = physio(bpp_strategy, 0, dev, B_WRITE, bpp_minphys, uio_p);
clean_cv_and_return:
	mutex_enter(&lp->bpp_mutex);
	cv_signal(&lp->wr_cv);
	mutex_exit(&lp->bpp_mutex);
	return error;
}

static void bpp_minphys(struct buf *bp) {
	minphys(bp);
	if (bp->b_bcount > 0x1000000) bp->b_bcount = 0x1000000;
}

static void check_for_active_pins(int ins)
{
	struct bpp_unit *lp;
	struct bpp_error_status *ep;
	volatile struct bpp_regs *regs;

	lp = ddi_get_soft_state(bpp_state_head, ins);
	ep = &lp->error_stat;
	regs = lp->bpp_regs_p;

	if (((regs->in_pins & BPP_ERR_PIN) && (regs->int_cntl & BPP_ERR_IRP)) ||
		((~regs->in_pins & BPP_ERR_PIN) && (~regs->int_cntl & BPP_ERR_IRP)))
	{
		ep->pin_status |= BPP_ERR_ERR;
	}

	if ( ((regs->in_pins & BPP_SLCT_PIN) && (regs->int_cntl & BPP_SLCT_IRP)) ||
		((~regs->in_pins & BPP_SLCT_PIN) && (~regs->int_cntl & BPP_SLCT_IRP)) )
	{
		ep->pin_status |= BPP_SLCT_ERR;
	}

	if ( ((regs->in_pins & BPP_PE_PIN) && (regs->int_cntl & BPP_PE_IRP)) ||
		((~regs->in_pins & BPP_PE_PIN) && (~regs->int_cntl & BPP_PE_IRP)) )
	{
		ep->pin_status |= BPP_PE_ERR;
	}
}

/*
 */
static int bpp_strategy(struct buf *bp)
{
	int ins;
	struct bpp_unit *lp;
	struct bpp_transfer_parms *tpp;
	volatile struct bpp_regs *regs;
	ddi_dma_cookie_t cookb;
	int dma_mode;
	int timeo;

	ins = getminor(bp->b_edev);
	lp = ddi_get_soft_state(bpp_state_head, ins);
	mutex_enter(&lp->bpp_mutex);
	tpp = &lp->transfer_parms;
	regs = lp->bpp_regs_p;

	if (lp->unit_suspended) {
		mutex_exit(&lp->bpp_mutex);
		ddi_dev_is_needed(lp->dip, 0 /*component*/, 1 /*level*/);
		mutex_enter(&lp->bpp_mutex);
	}

	lp->bpp_buffer = bp;
	lp->error_stat = bpp_default_error_stat;
	if (bp->b_flags & B_READ) {
		dma_mode = DDI_DMA_READ;
	} else if (bp->b_flags & B_WRITE) {
		dma_mode = DDI_DMA_WRITE;
	} else {
		cmn_err(CE_NOTE, "ERROR: bpp%d: b_flags confusion", ins);
		dma_mode = DDI_DMA_WRITE;
		goto bailout;
	}

	/* XXX We have to drop mutex here according to drivers examples. */
	if (ddi_dma_buf_setup(lp->dip, bp, dma_mode, DDI_DMA_SLEEP, 0,
					&bpp_limits, &lp->bpp_dma_handle) != 0)
	{
		cmn_err(CE_NOTE, "ERROR: bpp%d: ddi_dma_buf_setup failed mapping", ins);
		goto bailout;
	}
	if (ddi_dma_htoc(lp->bpp_dma_handle, 0, &cookb) != DDI_SUCCESS) {
		cmn_err(CE_NOTE, "ERROR: bpp%d: ddi_dma_htoc failed to fill in DMA cookie!", ins);
		goto bailout;
	}

	/* cmn_err(CE_NOTE, "bpp%d: addr=%08x%08x[%d]", ins, cookb._dmu._dmac_la[0], cookb._dmu._dmac_la[1], bp->b_bcount); */
	regs->dma_addr = cookb.dmac_address;
	regs->dma_bcnt = bp->b_bcount;
	regs->dma_csr |= BPP_ENABLE_BCNT;
	regs->dma_csr &= ~BPP_TC_INTR_DISABLE;
	regs->dma_csr |= BPP_ENABLE_DMA;

	if (bp->b_flags & B_READ) {
		timeo = tpp->read_timeout;
	} else {
		if ((lp->openflags & O_RDONLY) == 0 || (lp->openflags & O_WRONLY) == 0) {
			regs->int_cntl |= (BPP_ERR_IRQ_EN|BPP_SLCT_IRQ_EN|BPP_PE_IRQ_EN);
		}
		timeo = tpp->write_timeout;
	}

	lp->bpp_transfer_timeout_ident =
		timeout(bpp_transfer_timeout,
			(caddr_t)ins, drv_usectohz(timeo * 1000000));
	lp->timeouts |= TRANSFER_TIMEOUT;

	mutex_exit(&lp->bpp_mutex);
	return 0;

bailout:
	bp->b_flags |= B_ERROR;
	bp->b_error = EIO;
	mutex_exit(&lp->bpp_mutex);
	return 0;
}

static int bpp_ioctl(dev_t dev, int cmd, int arg, int mode,
          cred_t *cred_p, int *rval_p)
{
	int ins;
	struct bpp_unit *lp;
	struct bpp_error_status *esp;
	volatile struct bpp_regs *regs;
	struct bpp_transfer_parms parbuf;
	struct bpp_pins pinbuf;
	int rc = 0;
	enum handshake_t w_handshake, r_handshake;
	int r_rc = 0, w_rc = 0;

	ins = getminor(dev);
	lp = ddi_get_soft_state(bpp_state_head, ins);
	esp = &lp->error_stat;

	/* Taking some info from *lp before mutex_enter()? */
	regs = lp->bpp_regs_p;
	w_handshake = lp->transfer_parms.write_handshake;
	r_handshake = lp->transfer_parms.read_handshake;

	if (lp->unit_suspended) {
		ddi_dev_is_needed(lp->dip, 0 /*component*/, 1 /*level*/);
	}

	switch (cmd) {
	case BPPIOC_SETPARMS:	/* 126c  _IOW('b', 1, struct bpp_transfer_parms) */
		mutex_enter(&lp->bpp_mutex);
		copyin((caddr_t)arg, (caddr_t) &parbuf, sizeof(struct bpp_transfer_parms));
		if (mode & FREAD) {
			r_rc = (int) check_read_params(&parbuf, ins, mode);
		}
		if (mode & FWRITE) {
			w_rc = (int) check_write_params(&parbuf, ins, mode);
		}
		if (r_rc != 0 || w_rc != 0) return EINVAL;
		lp->transfer_parms = parbuf;
		mutex_exit(&lp->bpp_mutex);
		return rc;

	case BPPIOC_GETPARMS:	/* 1310  _IOR('b', 2, struct bpp_transfer_parms) */
		mutex_enter(&lp->bpp_mutex);
		copyout((caddr_t) &lp->transfer_parms, (caddr_t)arg, sizeof(struct bpp_transfer_parms));
		mutex_exit(&lp->bpp_mutex);
		return rc;

	case BPPIOC_SETOUTPINS:	/* 1338  _IOW('b', 3, struct bpp_pins) */
		mutex_enter(&lp->bpp_mutex);
		copyin((caddr_t)arg, (caddr_t) &pinbuf, sizeof(struct bpp_pins));
		if (mode & FREAD) {
			r_rc = check_read_pins(&pinbuf, mode, ins, r_handshake);
			if (r_rc == 0) {
				lp->pins = pinbuf;
			}
		}
		if (mode & FWRITE) {
			w_rc = check_write_pins(&pinbuf, mode, ins, w_handshake);
			if (w_rc == 0) {
				lp->pins = pinbuf;
			}
		}
		if (r_rc != 0 || w_rc != 0) return EINVAL;
		regs->out_pins = lp->pins.output_reg_pins;
		regs->in_pins = lp->pins.input_reg_pins;
		mutex_exit(&lp->bpp_mutex);
		return rc;

	case BPPIOC_GETOUTPINS:	/* 1410  _IOR('b', 4, struct bpp_pins) */
		mutex_enter(&lp->bpp_mutex);
		/* Yes, copyout, than read_outpins! */
		copyout((caddr_t) &lp->pins, (caddr_t)arg, sizeof(struct bpp_pins));
		read_outpins(ins, mode, w_handshake);
		mutex_exit(&lp->bpp_mutex);
		return rc;

	case BPPIOC_GETERR:	/* 1448   _IOR('b', 5, struct bpp_error_status) */
		mutex_enter(&lp->bpp_mutex);
		copyout((caddr_t) esp, (caddr_t)arg, sizeof(struct bpp_error_status));
		mutex_exit(&lp->bpp_mutex);
		return rc;

	case BPPIOC_TESTIO:	/* 1470   _IO('b', 6) */
		mutex_enter(&lp->bpp_mutex);
		*esp = bpp_default_error_stat;
		check_for_active_pins(ins);
		if (esp->pin_status == 0) rc = EIO;
		mutex_exit(&lp->bpp_mutex);
		return rc;

	default:
		return ENOTTY;
	}
}

/*
 */
static u_int
bpp_intr(caddr_t instance)
{
	struct bpp_unit *lp;
	volatile struct bpp_regs *regs;
	struct bpp_error_status *esp;
	u_int serviced = DDI_INTR_UNCLAIMED;
	struct buf *bp;

	lp = ddi_get_soft_state(bpp_state_head, (int)instance);
	mutex_enter(&lp->bpp_mutex);
	regs = lp->bpp_regs_p;
	esp = &lp->error_stat;

	if ((regs->dma_csr & BPP_INT_PEND) != 0 ||
	    (regs->dma_csr & BPP_ERR_PEND) != 0)
	{
		bp = lp->bpp_buffer;
		serviced = DDI_INTR_CLAIMED;

		if ((regs->dma_csr & BPP_ERR_PEND) != 0) {
			if ((regs->dma_csr & BPP_ENABLE_DMA) != 0) {
				bpp_transfer_failed((int)instance);
				mutex_exit(&lp->bpp_mutex);
				return DDI_INTR_CLAIMED;
			}
			bp->b_resid = bp->b_bcount;
			bp->b_flags |= B_ERROR;
			esp->bus_error = 1;
			bp->b_error = EIO;  /* =5 */
			while (((volatile struct buf*)bp)->b_flags & B_READ) {}
			bp->b_flags |= B_PHYS;
		}

		if ((regs->dma_csr & BPP_INT_PEND) != 0) {
			if ((regs->dma_csr & BPP_TERMINAL_CNT) != 0 &&
				(regs->dma_csr & BPP_TC_INTR_DISABLE) == 0)
			{
				regs->dma_csr |= BPP_TC_INTR_DISABLE;
				regs->dma_csr &= ~BPP_ENABLE_BCNT;
				regs->dma_csr |= BPP_TERMINAL_CNT;
				regs->dma_csr &= ~BPP_ENABLE_DMA;
				bp->b_resid = lp->transfer_remainder;
				regs->int_cntl &= ~(BPP_PE_IRQ_EN|BPP_SLCT_IRQ_EN|BPP_ERR_IRQ_EN);
			}

			if ((regs->int_cntl & BPP_ERR_IRQ) != 0 &&
			    (regs->int_cntl & BPP_ERR_IRQ_EN) != 0)
			{
				esp->pin_status |= BPP_ERR_PIN;
				if ((regs->dma_csr & BPP_ENABLE_DMA) != 0) {
					bpp_transfer_failed((int)instance);
				}
				regs->int_cntl |= BPP_ERR_IRQ;
			}
			if ((regs->int_cntl & BPP_SLCT_IRQ) != 0 &&
			    (regs->int_cntl & BPP_SLCT_IRQ_EN) != 0)
			{
				esp->pin_status |= BPP_SLCT_PIN;
				if ((regs->dma_csr & BPP_ENABLE_DMA) != 0) {
					bpp_transfer_failed((int)instance);
				}
				regs->int_cntl |= BPP_SLCT_IRQ;
			}
			if ((regs->int_cntl & BPP_PE_IRQ) != 0 &&
			    (regs->int_cntl & BPP_PE_IRQ_EN) != 0)
			{
				esp->pin_status |= BPP_PE_PIN;
				if ((regs->dma_csr & BPP_ENABLE_DMA) != 0) {
					bpp_transfer_failed((int)instance);
				}
				regs->int_cntl |= BPP_PE_IRQ;
			}
			if ((regs->int_cntl & BPP_BUSY_IRQ) != 0 &&
			    (regs->int_cntl & BPP_BUSY_IRQ_EN) != 0)
			{
				regs->int_cntl |= BPP_BUSY_IRQ;
			}
			if ((regs->int_cntl & BPP_ACK_IRQ) != 0 &&
			    (regs->int_cntl & BPP_ACK_IRQ_EN) != 0)
			{
				regs->int_cntl |= BPP_ACK_IRQ;
			}
			if ((regs->int_cntl & BPP_DS_IRQ) != 0 &&
			    (regs->int_cntl & BPP_DS_IRQ_EN) != 0)
			{
				regs->int_cntl |= BPP_DS_IRQ;
			}
		}

		untimeout(lp->bpp_transfer_timeout_ident);
		lp->timeouts &= ~TRANSFER_TIMEOUT;
		ddi_dma_free(lp->bpp_dma_handle);
		biodone(bp);
	}

	mutex_exit(&lp->bpp_mutex);
	return serviced;
}

/*
 * The bpp_transfer_failed locks not the bpp_mutex. Call it from under lock only.
 */
static void
bpp_transfer_failed(int instance)
{
	struct bpp_unit *lp;
	volatile struct bpp_regs *regs;
	struct buf *bp;

	lp = ddi_get_soft_state(bpp_state_head, instance);
	bp = lp->bpp_buffer;
	regs = lp->bpp_regs_p;

	regs->dma_csr &= ~BPP_ENABLE_DMA;

	if ((regs->op_config & BPP_IDLE) == 0) {
		regs->dma_csr &= ~BPP_ENABLE_BCNT;
		regs->op_config |= BPP_SRST;
		regs->op_config &= ~BPP_SRST;
		bp->b_resid = (regs->dma_bcnt - 1) + lp->transfer_remainder;
	} else {
		regs->dma_csr &= ~BPP_ENABLE_BCNT;
		bp->b_resid = regs->dma_bcnt + lp->transfer_remainder;
	}

	regs->dma_csr |= BPP_FLUSH;
	regs->dma_bcnt = 0;
	regs->dma_csr |= BPP_TC_INTR_DISABLE;
	regs->int_cntl &= ~(BPP_PE_IRQ_EN|BPP_SLCT_IRQ_EN|BPP_ERR_IRQ_EN);

	check_for_active_pins(instance);
}

static void
bpp_transfer_timeout(caddr_t arg)
{
	struct bpp_unit *lp;
	volatile struct bpp_regs *regs;
	struct buf *bp;

	lp = ddi_get_soft_state(bpp_state_head, (int)arg);
	mutex_enter(&lp->bpp_mutex);
	regs = lp->bpp_regs_p;

	bp = lp->bpp_buffer;
	if ((regs->trans_cntl & BPP_DIRECTION) != 0) {
		bp->b_resid = regs->dma_bcnt + lp->transfer_remainder;
		regs->dma_bcnt = 0;
		regs->dma_csr |= BPP_TC_INTR_DISABLE;
		regs->dma_csr &= ~BPP_ENABLE_BCNT;
		regs->dma_csr &= ~BPP_ENABLE_DMA;
		regs->int_cntl &= ~(BPP_PE_IRQ_EN|BPP_SLCT_IRQ_EN|BPP_ERR_IRQ_EN);
	} else {
		if ((regs->trans_cntl & BPP_DIRECTION) == 0) {
			bpp_transfer_failed((int)arg);
			if (bp->b_resid == 0) {
				bp->b_flags |= B_ERROR;
				bp->b_error = EIO;
			}
		}
	}
	lp->timeouts &= ~TRANSFER_TIMEOUT;
	lp->error_stat.timeout_occurred = 1;
	ddi_dma_free(lp->bpp_dma_handle);	/* Check rc here? */
	mutex_exit(&lp->bpp_mutex);
	biodone(bp);
}

/*
 */
static void
set_dss_dsw(int instance, int read)
{
	struct bpp_unit *lp;
	struct bpp_transfer_parms *tpp;
	unsigned dss, dsw;

	lp = ddi_get_soft_state(bpp_state_head, instance);
	tpp = &lp->transfer_parms;
	if (read) {
		dss = tpp->read_setup_time / lp->sbus_clock_cycle;
		if (tpp->read_setup_time % lp->sbus_clock_cycle != 0) dss++;
		dsw = tpp->read_strobe_width / lp->sbus_clock_cycle;
		if (tpp->read_strobe_width % lp->sbus_clock_cycle != 0) dsw++;
	} else {
		dss = tpp->write_setup_time / lp->sbus_clock_cycle;
		if (tpp->write_setup_time % lp->sbus_clock_cycle != 0) dss++;
		dsw = tpp->write_strobe_width / lp->sbus_clock_cycle;
		if (tpp->write_strobe_width % lp->sbus_clock_cycle != 0) dsw++;
	}

	/* Wrong masks!! */
	lp->bpp_regs_p->hw_config = (dss & 0xff) | ((dsw & 0xff) << 8);
}

/*
 */
static short
check_write_params(struct bpp_transfer_parms *pp, int instance, int mode)
{
 	/* Why static ?! */
	static unsigned data_e0, data_e4;

	struct bpp_unit *lp;

	lp = ddi_get_soft_state(bpp_state_head, instance);

	/*
	 * All this mess is true, checked many times.
	 */
	if (pp->write_handshake < 1) return EINVAL;
	if (pp->write_handshake > 10) return EINVAL;
	if (pp->write_handshake > 3) {
		if (pp->write_handshake < 9) return EINVAL;
	}
	/* ????? !!! */
	if ((mode & FREAD) != 0) {
		if (pp->write_handshake > 3) return EINVAL;
	}
	if (pp->write_handshake > 3) return EINVAL;

	data_e0 = (0 - lp->sbus_clock_cycle) + (lp->sbus_clock_cycle << 7);  /* cycle*127 */
	data_e4 = (0 - lp->sbus_clock_cycle) + (lp->sbus_clock_cycle << 7);

	if (pp->write_setup_time < 0) return EINVAL;
	if (pp->write_setup_time > data_e0) return EINVAL;
	if (pp->write_strobe_width < 0) return EINVAL;
	if (pp->write_strobe_width > data_e4) return EINVAL;

	if (pp->write_timeout < 0) return EINVAL;
	if (pp->write_timeout > 1800000) return EINVAL;

	return 0;
}

/*
 */
static short
check_read_params(struct bpp_transfer_parms *pp, int instance, int mode)
{
	static unsigned data_e8, data_ec;

	struct bpp_unit *lp;

	lp = ddi_get_soft_state(bpp_state_head, instance);

	if (pp->read_handshake < 1) return EINVAL;
	if (pp->read_handshake > 8) return EINVAL;

	data_e8 = lp->sbus_clock_cycle * 127;
	data_ec = lp->sbus_clock_cycle * 127;
	if (pp->read_setup_time < 0) return EINVAL;
	if (pp->read_setup_time > data_e8) return EINVAL;
	if (pp->read_strobe_width < 0) return EINVAL;
	if (pp->read_strobe_width > data_ec) return EINVAL;

	if (pp->read_timeout < 0) return EINVAL;
	if (pp->read_timeout > 1800000) return EINVAL;

	return 0;
}

static int
check_read_pins(struct bpp_pins *pp, int mode, int instance,
	enum handshake_t handshake)
{
	struct bpp_unit *lp;

	lp = ddi_get_soft_state(bpp_state_head, instance);
	/* lp not used, true. */

	if ((pp->output_reg_pins & ~BPP_ALL_OUT_PINS) != 0) return EINVAL;
	if ((pp->input_reg_pins & ~BPP_ALL_IN_PINS) != 0) return EINVAL;
	return 0;
}

static int
check_write_pins(struct bpp_pins *pp, int mode, int instance,
	enum handshake_t handshake)

{
	struct bpp_unit *lp;

	lp = ddi_get_soft_state(bpp_state_head, instance);

	if ((pp->output_reg_pins & ~BPP_ALL_OUT_PINS) != 0) return EINVAL;
	if ((pp->input_reg_pins & ~BPP_ALL_IN_PINS) != 0) return EINVAL;
	if (handshake > 3) return EINVAL;
	return 0;
}

/*
 */
static void
read_outpins(int instance, int mode, int handshake)
{
	struct bpp_unit *lp;
	volatile struct bpp_regs *regs;

	lp = ddi_get_soft_state(bpp_state_head, instance);
	if (mode & FWRITE) {
		regs = lp->bpp_regs_p;
		lp->pins.output_reg_pins |= regs->out_pins & (BPP_SLCTIN_PIN|BPP_AFX_PIN|BPP_INIT_PIN);
	}
}

static int
check_bpp_registers(int instance)
{
	long lbuf;
	short sbuf;
	char cbuf;
	struct bpp_unit *lp;

	lp = ddi_get_soft_state(bpp_state_head, instance);

	if (ddi_peekl(lp->dip, (long*) &lp->bpp_regs_p->dma_csr, &lbuf) != 0) return 1;
	if (ddi_peekl(lp->dip, (long*) &lp->bpp_regs_p->dma_addr, &lbuf) != 0) return 1;
	if (ddi_peekl(lp->dip, (long*) &lp->bpp_regs_p->dma_bcnt, &lbuf) != 0) return 1;
	if (ddi_peeks(lp->dip, (short*) &lp->bpp_regs_p->hw_config, &sbuf) != 0) return 1;
	if (ddi_pokes(lp->dip, (short*) &lp->bpp_regs_p->hw_config, sbuf) != 0) return 1;
	if (ddi_peeks(lp->dip, (short*) &lp->bpp_regs_p->op_config, &sbuf) != 0) return 1;
	if (ddi_peeks(lp->dip, (short*) &lp->bpp_regs_p->int_cntl, &sbuf) != 0) return 1;
	if (ddi_peekc(lp->dip, (char*) &lp->bpp_regs_p->data, &cbuf) != 0) return 1;
	if (ddi_peekc(lp->dip, (char*) &lp->bpp_regs_p->trans_cntl, &cbuf) != 0) return 1;
	if (ddi_peekc(lp->dip, (char*) &lp->bpp_regs_p->out_pins, &cbuf) != 0) return 1;
	if (ddi_peekc(lp->dip, (char*) &lp->bpp_regs_p->in_pins, &cbuf) != 0) return 1;
	return 0;
}
